"""ЯДРО"""

from dataclasses import fields, is_dataclass
from itertools import chain
from typing import Iterable, Mapping

from ..logger.core import log


def pretty_print(obj, indent=4, title='PRETTY_PRINT', m2d: bool = False, outputter=log):
    """
    Pretty prints a (possibly deeply-nested) dataclass.
    Each new block will be indented by `indent` spaces (default is 4).
    """
    obj = model_to_dict(obj) if m2d else obj
    outputter(title)
    outputter(stringify(obj, indent))


def model_to_dict(instance, fields=None, exclude=None):
    """
    Return a dict containing the data in ``instance`` suitable for passing as
    a Form's ``initial`` keyword argument.

    ``fields`` is an optional list of field names. If provided, return only the
    named.

    ``exclude`` is an optional list of field names. If provided, exclude the
    named from the returned dict, even if they are listed in the ``fields``
    argument.
    """
    opts = instance._meta
    data = {}
    for f in chain(opts.concrete_fields, opts.private_fields, opts.many_to_many):
        if not getattr(f, "editable", False):
            continue
        if fields is not None and f.name not in fields:
            continue
        if exclude and f.name in exclude:
            continue
        data[f.name] = f.value_from_object(instance)
    return data


def stringify(obj, indent=4, _indents=0):
    """Stringify object to str"""
    if isinstance(obj, str):
        return f"'{obj}'"

    if not is_dataclass(obj) and not isinstance(obj, (Mapping, Iterable)):
        return str(obj)

    this_indent = indent * _indents * ' '
    next_indent = indent * (_indents + 1) * ' '
    start, end = f'{type(obj).__name__}(', ')'  # dicts, lists, and tuples will re-assign this

    if is_dataclass(obj):
        body = '\n'.join(
            f'{next_indent}{field.name}='
            f'{stringify(getattr(obj, field.name), indent, _indents + 1)},' for field in fields(obj)
        )

    elif isinstance(obj, Mapping):
        if isinstance(obj, dict):
            start, end = '{}'

        body = '\n'.join(
            f'{next_indent}{stringify(key, indent, _indents + 1)}: '
            f'{stringify(value, indent, _indents + 1)},' for key, value in obj.items()
        )

    else:  # is Iterable
        if isinstance(obj, list):
            start, end = '[]'
        elif isinstance(obj, tuple):
            start = '('

        body = '\n'.join(
            f'{next_indent}{stringify(item, indent, _indents + 1)},' for item in obj
        )

    return f'{start}\n{body}\n{this_indent}{end}'



import json
import re
import shutil
from itertools import islice
from pathlib import Path
from typing import Any, Final, Optional

import tiktoken

from .. import ColumnStrings, logerr, pretty_print
from ..configs.system import AsisTobe
from ..logger import log, logsuc
from ..reader_controller.exceptions import EmptyIdfyNotNullDictException
from ..reader_controller.types import IdfyGoods, IndexedIdfyGoods

print = log


class Tools:
    # Глобальный паттерн всех Unicode-пробелов (включая zero-width)
    _UNICODE_SPACES_RE = re.compile(
        r"[\u0009\u000A\u000B\u000C\u000D\u0020\u0085\u00A0\u1680"
        r"\u180E\u2000-\u200A\u2028\u2029\u202F\u205F\u3000\u200B\u2060\uFEFF]+"
    )

    @staticmethod
    def semicolonizer(string: str) -> str:
        string = string.replace(",", ", ")
        return string.replace("  ", " ")

    @staticmethod
    def extract_digits_list(s: str) -> list[str]:
        """
        Извлекает все числа (последовательности цифр) из строки.
        Возвращает список строк (не int, чтобы сохранить ведущие нули, если нужно).
        """
        return re.findall(r'\d+', s)

    @staticmethod
    def get_dry_brand(brand: str):
        allowed_chars = (
            'abcdefghijklmnopqrstuvwxyz'
            'абвгдеёжзийклмнопрстуфхцчшщъыьэюя'
            '0123456789&!-\''
        )
        return Tools.get_dry_string(brand, allowed_chars)

    @staticmethod
    def get_dry_string(input_string: str, allows: Optional[str] = None) -> str:
        """Высушивает строку, оставляя только буквы и цифры в lower-формате"""
        allowed_chars = (
            'abcdefghijklmnopqrstuvwxyz'
            'абвгдеёжзийклмнопрстуфхцчшщъыьэюя'
            '0123456789'
        )
        if allows is not None:
            allowed_chars = allows
        return ''.join(
            c for c in input_string.lower()
            if c in allowed_chars
        )

    @staticmethod
    def get_dry_string_list(input_string: str) -> list[str]:
        """Высушивает строку, оставляя только буквы и цифры в lower-формате в списке"""
        result: list[str] = []
        for word in input_string.split():
            result.append(Tools.get_dry_string(word))
        return result

    @staticmethod
    def get_all_fields(fields: list[str]) -> list[str]:
        """
        Расширяет список полей, добавляя суффиксы _asis и _tobe
        :param fields: список базовых полей
        :return: список с расширенными названиями
        """
        if not fields:
            raise ValueError("Список полей не может быть пустым")
        extended_fields = []
        for field in fields:
            extended_fields.append(f"{field}_asis")
            extended_fields.append(f"{field}_tobe")
        return extended_fields

    @staticmethod
    def get_relay(fields: list[str]) -> dict[str, str]:
        """Генерирует словарь с asis->tobe связкой из fields"""
        if not len(fields):
            raise IndexError("СПИСОК ПОЛЕЙ ДЛЯ ФИНАЛИЗАЦИИ НЕ МОЖЕТ БЫТЬ ПУСТЫМ")
        result: dict[str, str] = {}
        for field in fields:
            result[f"{field}_asis"] = f"{field}_tobe"
        return result

    @staticmethod
    def normalize_spaces(text: str) -> str:
        """
        Заменяет все типы пробельных символов Unicode на обычный пробел.
        """
        return Tools._UNICODE_SPACES_RE.sub(' ', text)

    @staticmethod
    def string_stripper(text: str) -> str:
        """
        - Удаляет ведущие/конечные пробелы
        - Заменяет все невидимые и странные пробелы на обычные
        - Схлопывает многократные пробелы в один
        """
        if not text:
            return ""
        text = Tools.normalize_spaces(text)
        text = text.strip()
        text = re.sub(r' {2,}', ' ', text)
        return text.strip()

    @staticmethod
    def clean_field(string: str) -> str:
        """Комплексная очистка"""
        string = Tools.string_stripper(string)
        string = Tools.trim_spec(string)
        return Tools.string_stripper(string)

    @staticmethod
    def polysplit(
        text: str,
        separators: Optional[list[str]] = None,
        no_empty: bool = False
    ) -> list[str]:
        """
        Разбивает строку по нескольким разделителям с возможностью удаления пустых строк.

        :param text: Исходная строка
        :param separators: Список разделителей
        :param no_empty: Если True (по умолчанию), удаляет пустые строки из результата
        :return: Список частей строки
        """
        if separators is None:
            separators = ["/", "\\", "|", "-", "+", " "]
        if not text:
            return []

        result = [text]

        for sep in separators:
            temp = []
            for part in result:
                temp.extend(part.split(sep))
            result = temp

        if no_empty:
            return [part for part in result if part]  # Фильтруем пустые строки
        return result  # Возвращаем как есть

    @staticmethod
    def get_words(string: str, use_clean_strip: bool = False) -> list[str]:
        """
        Получает из строки список слов по пробелу

        use_clean_strip - если задан, возьмёт множество других символов
        """
        string = Tools.normalize_spaces(string)
        string_list = string.split(" ")
        if use_clean_strip:
            # УБРАЛИ "!"
            return [word.strip('.,?:;"\'()-_/|*<>#%^&+=~ ') for word in string_list]
        return string_list

    @staticmethod
    def get_lower_words(words: list[str]) -> list[str]:
        """Берет список строк, делает их lower"""
        return [word.lower() for word in words]

    @staticmethod
    def get_most_longer_word_of_list(words_list):
        """
        Метод, который выбирает самое длинное название из списка.
        Если таких несколько - сортируем по алфавиту.

        :param words_list: Список слов
        :return: Самое длинное слово или первое из самых длинных, если их несколько
        """
        if not words_list:
            return None

        # Находим максимальную длину слова
        max_length = max(len(word) for word in words_list)

        # Собираем список всех слов максимальной длины
        longest_words = [word for word in words_list if len(word) == max_length]

        # Если одно слово, возвращаем его
        if len(longest_words) == 1:
            return longest_words[0]
        else:
            # Если несколько слов одной длины, сортируем по алфавиту и берем первое
            return sorted(longest_words)[0]

    @staticmethod
    def get_first_of_dict(dictionary: dict, key: bool = False) -> Optional[Any]:
        """Возвращает первый элемент ключа"""
        priority_key = next(iter(dictionary), None)
        if priority_key is None:
            return None
        return priority_key if key else dictionary[priority_key]

    @staticmethod
    def trim_spec(word: str, is_brand: bool = False) -> str:
        """
        Очищает строку от спецсимволов по краям.
        При is_brand=True:
          - Если строка — одно слово в скобках (например, "(Samsung)") → удаляет скобки.
          - Если слов несколько или скобки не по краям → оставляет как есть.

        Args:
            word (str): Исходная строка.
            is_brand (bool): Флаг, указывающий, что строка — бренд (требует особой обработки скобок).

        Returns:
            str: Очищенная строка.
        """
        if not word or not isinstance(word, str):
            return word

        word = word.strip()

        # Базовая очистка краёв: удаляем всё, кроме букв, цифр, дефиса, апострофа, скобок
        cleaned = re.sub(r'^[^\w\(\)\[\]\{\}\'-]+|[^\w\(\)\[\]\{\}\'-]+$', '', word)

        if not is_brand:
            return cleaned

        # === Логика для брендов ===

        # Если строка уже не в скобках — просто возвращаем
        if not (cleaned.startswith('(') and cleaned.endswith(')')):
            return cleaned

        # Проверяем, является ли содержимое "одним словом"
        inner = cleaned[1:-1].strip()  # убираем скобки и пробелы

        # Разбиваем на "слова" (последовательности букв/цифр)
        inner_words = re.findall(r'[\w]+', inner, re.UNICODE)

        # Если внутри 2+ слов — оставляем скобки
        if len(inner_words) >= 2:
            return cleaned

        # Если внутри 1 слово — удаляем внешние скобки
        if len(inner_words) == 1:
            return inner

        # Если внутри 0 слов (пусто или только символы) — возвращаем без скобок
        return inner  # может быть пустым

    @staticmethod
    def get_most_popular(variants_list: list[str], variants_registry: dict[str, int]) -> list[str]:
        """
        Выбирает все самые популярные варианты из variants_list
        Популярность содержится в variants_registry в виде словаря: {
            "бренд1": <количество_повторений_в_документе>,
            "бренд2": <количество_повторений_в_документе>,
        }
        Если у одного из вариантов нет повторений в variants_registry, то приравнять популярность к 0
        Если у нескольких вариантов одинаковая максимальная популярность, вернуть все такие варианты
        :param variants_list: список брендов для анализа
        :param variants_registry: словарь популярности брендов
        :return: список самых популярных брендов
        """
        # Сначала найдем максимальную частоту
        max_frequency = -1
        for brand in variants_list:
            frequency = variants_registry.get(Tools.get_dry_brand(brand), 0)
            if frequency > max_frequency:
                max_frequency = frequency

        # print(f"{max_frequency=}")

        # Теперь соберем все бренды с максимальной частотой
        popular_brands = []
        for brand in variants_list:
            frequency = variants_registry.get(Tools.get_dry_brand(brand), 0)
            if frequency == max_frequency:
                popular_brands.append(brand)

        return popular_brands

    @staticmethod
    def get_threads_data_parts_by_dict(
        idfy_not_null_dict: IdfyGoods,
        thread_pks: list[int],
        is_use_small_chunks: bool = False,
        small_chunk_size: int = 10,
    ) -> IndexedIdfyGoods:
        """
        Разбивает словарь данных на части по списку thread_pks.
        Если данных < 10 — все данные идут в первый поток (по списку).
        Иначе — распределяются равномерно.

        :param idfy_not_null_dict: Исходный словарь данных (ключ: int, значение: dict)
        :param thread_pks: Список pk потоков (например, из .values_list('pk', flat=True))
        :param is_use_small_chunks: Ограничить объём данных (для тестов)
        :param small_chunk_size: Макс. кол-во записей, если is_use_small_chunks=True
        :return: dict[thread_pk: dict[int, XLSGood]] — ключ — pk потока
        :raises ValueError: Если thread_pks пуст
        :raises EmptyIdfyNotNullDictException: Если нет данных
        """
        log("Генерирую словарь с данными для потоков")

        if not thread_pks:
            raise ValueError("Список thread_pks пуст")

        if not idfy_not_null_dict:
            raise EmptyIdfyNotNullDictException("Нет данных")

        # Подготовка данных
        items = list(idfy_not_null_dict.items())
        if is_use_small_chunks:
            items = items[:small_chunk_size]
        total_items = len(items)
        total_threads = len(thread_pks)

        result: IndexedIdfyGoods = {}

        # 🔹 Логика: если данных мало — всё в первый поток
        if total_items < 10:
            first_pk = thread_pks[0]
            log(f"Мало данных ({total_items} < 10). Все данные в поток {first_pk}")
            result[first_pk] = dict(items)
            # Остальные потоки — пустые
            for pk in thread_pks[1:]:
                result[pk] = {}
        else:
            # Обычное равномерное распределение
            base_size, remainder = divmod(total_items, total_threads)
            index = 0

            for pk in thread_pks:
                chunk_size = base_size + (1 if remainder > 0 else 0)
                chunk = dict(items[index:index + chunk_size])
                result[pk] = chunk
                index += chunk_size
                remainder -= 1

        avg_per_thread = total_items // total_threads if total_threads else 0
        logsuc(f"Генерация словаря завершена. ~ записей в потоке: {avg_per_thread}")
        return result

    @staticmethod
    def get_dict_filtered(
        data: dict[Any, dict[str, Any]],
        include_by_rules: list[dict[str, Any]],
        strict: bool = False
    ) -> dict[Any, dict[str, Any]]:
        """
        Фильтрует словарь словарей по правилам.
        Поддерживает:
          - Несколько условий на одно поле → объединяются как OR
          - Условия на разные поля → объединяются как AND

        Пример:
            rules = [
                {'status': 'active'},
                {'status': 'pending'},
                {'category': 'premium'}
            ]
            → (status == 'active' OR status == 'pending') AND category == 'premium'

        :param data: словарь, где значения — словари
        :param include_by_rules: список фильтров (условия на поля)
        :param strict: если True, отсутствие ключа = ошибка; иначе — несовпадение
        :return: отфильтрованный словарь
        """
        if not isinstance(data, dict):
            raise ValueError("data должен быть словарём")
        if not isinstance(include_by_rules, list):
            raise ValueError("include_by_rules должен быть списком")

        # Если нет правил — возвращаем все подходящие элементы
        if not include_by_rules:
            return {k: v for k, v in data.items() if isinstance(v, dict)}

        # Группируем правила по ключам: field -> set(expected_values)
        conditions: dict[str, set[Any]] = {}
        for rule in include_by_rules:
            if not isinstance(rule, dict):
                continue
            for key, value in rule.items():
                if key not in conditions:
                    conditions[key] = set()
                conditions[key].add(value)

        result = {}

        for key, item in data.items():
            if not isinstance(item, dict):
                continue

            match = True
            for field, allowed_values in conditions.items():
                if field not in item:
                    # Ключ отсутствует
                    if strict:
                        match = False
                    else:
                        match = False  # значение не совпадает (ключа нет)
                    break

                actual_value = item[field]
                if actual_value not in allowed_values:
                    match = False
                    break  # не прошло условие по полю

            if match:
                result[key] = item

        return result

    @staticmethod
    def skip_n_rows(data_dict: IdfyGoods, n_of_skip_rows: int) -> IdfyGoods:
        """Пропуск указанного количества чанков"""
        return dict(list(data_dict.items())[n_of_skip_rows:])

    @staticmethod
    def no_repeats_of_list(lst: list) -> list:
        """
        Удаляет дубликаты из списка, сохраняя порядок элементов.

        :param lst: Исходный список
        :return: Список без дубликатов
        """
        seen = set()
        filtered = [x for x in lst if not (x in seen or seen.add(x))]
        return [x for x in filtered if x is not None]

    @staticmethod
    def find_original_substring(source: str, word: str) -> Optional[str]:
        """
        Находит подстроку в исходной строке, соответствующую образцу (с учетом регистра и оригинального написания).

        :param source: Исходная строка для поиска
        :param word: Подстрока-образец (может быть в нижнем регистре)
        :return: Найденная подстрока в оригинальном виде или None
        """
        # Экранируем спецсимволы в образце и разбиваем на слова
        pattern_parts = [re.escape(part) for part in word.split()]

        # Создаем регулярное выражение:
        # - Ищем слова образца в любом порядке
        # - Между словами могут быть другие символы
        # - Регистронезависимый поиск
        regex_pattern = r'(?i)(?=.*?\b{}\b)'.format(r'\b)(?=.*?\b'.join(pattern_parts))
        regex_pattern += r'([^\n]*?{}[^\n]*)'.format(r'[^\n]*?'.join(pattern_parts))

        match = re.search(regex_pattern, source)
        if not match:
            return None

        # Находим точную подстроку, содержащую все слова образца
        matched_text = match.group(1)

        # Уточняем границы найденного текста
        words = word.lower().split()
        start_pos = 0
        end_pos = len(matched_text)

        # Находим первое вхождение первого слова
        first_word_pattern = re.compile(re.escape(words[0]), re.IGNORECASE)
        first_match = first_word_pattern.search(matched_text)
        if first_match:
            start_pos = first_match.start()

        # Находим последнее вхождение последнего слова
        last_word_pattern = re.compile(re.escape(words[-1]), re.IGNORECASE)
        last_matches = list(last_word_pattern.finditer(matched_text))
        if last_matches:
            end_pos = last_matches[-1].end()

        # Возвращаем подстроку с оригинальным написанием
        return matched_text[start_pos:end_pos].strip(" ,.-")

    @staticmethod
    def get_normal_word_from_source_by_castrat_word(
        source: str,
        word: str
    ):
        """
        Возвращает нормальное слово из источника по индексу, найденному по кастрату в источнике
        :param source: example = "Стиральная машина AEG L7WBE68SI White"
        :param word: example = "white"
        :return: example = "White"
        """
        # debug: bool = False
        _debug: bool = True

        if _debug: logerr("<get_normal_word_from_source_by_castrat_word>")
        if _debug: print(f"{source=}")
        if _debug: print(f"{word=}")
        source_list: list[str] = source.split()
        if _debug: pretty_print(source_list, title=f"source_list", m2d=False)
        source_lower: str = source.lower()
        if _debug: print(f"{source_lower=}")

        source_lower_list: list[str] = source_lower.split()
        source_lower_list = [Tools.trim_spec(color) for color in source_lower_list]
        if _debug: pretty_print(source_lower_list, title=f"source_lower_list", m2d=False)

        index_main: int = 0
        for index, word_part in enumerate(source_lower_list):
            if word not in word_part:
                continue
            else:
                index_main = index
                break
        if _debug: print(f"{index_main=}")

        if not index_main:
            return None
        return source_list[index_main]

    @staticmethod
    def completely_nulled(fields: list[str], data_dict: dict) -> bool:
        """
        Возвращает True, если все указанные поля в data_dict равны строке "0".
        :param fields: список ключей для проверки
        :param data_dict: словарь с данными
        :return: True, если все указанные поля равны "0", иначе False
        """
        return all(data_dict.get(field) == "0" for field in fields)

    @staticmethod
    def get_stripped_words(words: list[str], chars: int = 3) -> list[str]:
        """
        Получает список слов words и сколько нужно оставить в начале слова chars,
        возвращается список обрезанных слов
        :param words:
        :param chars:
        :return:
        """
        if chars < 0:
            raise ValueError("Количество символов не может быть отрицательным")
        return [word[:chars] for word in words]

    @classmethod
    def get_stripped_data_by_fields(cls, data: dict, required_fields: list[str]):
        """
        Получает словарь с полями. Пробегается по полю из required_field. Чистит от пробелов
        :param data:
        :param required_fields:
        :return:
        """
        result: dict[str, str] = {}
        for field, value in data.items():
            if field not in required_fields:
                result[field] = value
                continue
            result[field] = cls.string_stripper(data[field])

        return result

    @staticmethod
    def get_desymbolization_string(string: str, allowed_symbols: list):
        """Оставляем буквы-цифры и разрешенные символы"""
        # Экранируем спецсимволы для безопасного включения в регулярку
        escaped = ''.join(re.escape(sym) for sym in allowed_symbols)
        # Регулярка оставляет A-z, А-я, 0-9 и разрешённые символы
        allowed_pattern = fr"[^a-zA-Zа-яА-ЯёЁ0-9{escaped}]"
        return re.sub(allowed_pattern, " ", string)

    @staticmethod
    def ireplace(
        string: str,
        substr: str,
        value: str = ""
    ) -> str:
        """
        Удаляет все вхождения подстроки `substring` из строки `string` без учёта регистра.
        Например: string="HelloWorldWorld", substring="world" → "Hello"
        """
        pattern = re.compile(re.escape(substr), re.IGNORECASE)
        return pattern.sub(value, string)

    @staticmethod
    def get_sliced_after(string: str, phrase: str) -> str:
        """
        Метод ищет в строке string начало вхождения фразы phrase.
        Если находит — возвращает часть строки ДО фразы.
        Если не находит — возвращает исходную строку.

        Пример:
            string="abc", phrase="b" → return "a"

        :param string: Исходная строка
        :param phrase: Фраза, после которой нужно обрезать
        :return: Обрезанная строка
        """
        string_lower = string.lower()
        phrase_lower = phrase.lower()

        index = string_lower.find(phrase_lower)
        if index != -1:
            return string[:index]
        return string  # если phrase не найдено — вернуть оригинал без изменений

    @staticmethod
    def ifound(
        string: str, substring: str, is_all: bool = False
    ) -> bool | list[int]:
        """
        Выполняет регистронезависимый поиск подстроки `substring` в строке `string`.

        :param string: Строка, в которой происходит поиск
        :param substring: Подстрока, которую ищем
        :param is_all: Если True — возвращает список всех индексов вхождения;
                       Если False — возвращает True/False в зависимости от наличия вхождения
        :return: bool или list[int]
        """
        string_lower = string.lower()
        substring_lower = substring.lower()

        if not is_all:
            return substring_lower in string_lower

        # Поиск всех вхождений
        indices = []
        start = 0
        while True:
            index = string_lower.find(substring_lower, start)
            if index == -1:
                break
            indices.append(index)
            start = index + 1  # Продолжаем искать со следующего символа

        return indices

    @staticmethod
    def try_to_int(data: Optional) -> bool:
        try:
            int(data)
            return True
        except (ValueError, TypeError):
            return False

    @staticmethod
    def get_all_files_from_dir(
        dir_path: Path,
        exts_list: list[str] | None = None,
        exclude_file_with_list: list[str] | None = None,
        sort_files: bool = True
    ) -> list[Path]:
        """
        Собирает и сортирует файлы из указанной директории

        :param dir_path: путь до папки
        :param exts_list: разрешенные расширения (без точки, например ['jpg', 'png'])
        :param exclude_file_with_list: исключить файлы, содержащие подстроки в имени
        :param sort_files: сортировать ли файлы по имени (по умолчанию True)
        :return: отсортированный список Path объектов файлов
        """
        exts_list = exts_list or ['*']
        exclude_file_with_list = exclude_file_with_list or ['~']

        if not dir_path.is_dir():
            raise NotADirectoryError(f"Путь не является директорией: {dir_path}")

        found_files: list[Path] = []

        for ext in exts_list:
            pattern = f"*.{ext}" if ext != '*' else "*"
            found_files.extend(
                file_path
                for file_path in dir_path.glob(pattern)
                if file_path.is_file()
                and not any(
                    exclude_str in file_path.name
                    for exclude_str in exclude_file_with_list
                )
            )

        # Удаляем дубликаты и сортируем при необходимости
        unique_files = list(set(found_files))
        return sorted(unique_files, key=lambda x: x.name) if sort_files else unique_files

    @staticmethod
    def sort_dict_by_keys(dictionary: dict, reverse=False):
        """
        Сортирует словарь по ключам.

        :param dictionary: Словарь для сортировки.
        :param reverse: Флаг для выбора направления сортировки (False - прямой порядок, True - обратный).
        :return: Отсортированный словарь.
        """
        sorted_items = sorted(dictionary.items(), key=lambda x: x[0], reverse=reverse)
        return dict(sorted_items)

    @staticmethod
    def sequential_combinations(
        words_list: list[str],
        use_dry: bool = True,
        get_string: bool = True,
        is_brand: bool = False
    ):
        """
        Составляет последовательные комбинации слов из списка и сортирует их по убыванию длины.
        Если get_string=True, возвращает строки вместо списков.

        :param is_brand:
        :param use_dry:
        :param words_list: Входной список слов
        :param get_string: Флаг, определяющий формат вывода (списки или строки)
        :return: Список последовательных комбинаций, отсортированных по убыванию длины
        """
        result = []
        length = len(words_list)

        # Генерация последовательных комбинаций
        for start_idx in range(length):
            for end_idx in range(start_idx + 1, length + 1):
                combo = words_list[start_idx:end_idx]
                result.append(combo)

        # Сортировка по убыванию длины
        result.sort(key=len, reverse=True)

        if get_string:
            result = [' '.join(combo) for combo in result]

        if use_dry:
            if is_brand:
                result = [Tools.get_dry_brand(combo) for combo in result]
            else:
                result = [Tools.get_dry_string(combo) for combo in result]

        return result

    @staticmethod
    def get_no_entity_no_brand_source_name(
        data_item: dict,
        no_entity: bool = True,
        no_brand: bool = True
    ) -> str:
        source_name: str = data_item.get(ColumnStrings.DATA_SOURCE_NAME)
        if no_entity:
            source_name = source_name.replace(data_item.get(ColumnStrings.DATA_ENTITY_ASIS), "")
        if no_brand:
            source_name = source_name.replace(data_item.get(ColumnStrings.DATA_BRAND_ASIS), "")
        return source_name.strip()

    @staticmethod
    def find_full_match(source_name: str, pattern: str) -> str:
        """
        Метод находит точное полное вхождение подстроки в исходной строке.
        Окончанием считается пробел, запятая, перенос строки или конец строки.

        :param source_name: Исходная строка
        :param pattern: Часть строки, по которой нужно искать
        :return: Полное вхождение искомой строки
        """
        pos = source_name.find(pattern)
        if pos == -1:
            return ""

        # Начинаем движение вправо от места нахождения паттерна
        end_pos = pos + len(pattern)
        while end_pos < len(source_name) and source_name[end_pos].isalnum():  # Пока буква или цифра
            end_pos += 1

        # Вырезаем найденное вхождение
        full_match = source_name[pos:end_pos].strip(", \n\r\t")
        return full_match

    @staticmethod
    def get_smallest(*lists):
        """Выбирает наименьшее ненулевое число из переданных списков.
        Если все значения нулевые или списки пусты, возвращает None."""
        # # Собираем все ненулевые элементы из всех списков
        # non_zero_values = []
        # for lst in lists:
        #     if lst:  # Проверяем, что список не пустой
        #         for value in lst:
        #             if value != 0:  # Игнорируем нули
        #                 non_zero_values.append(value)
        non_zero_values = [v for lst in lists if lst for v in lst if v != 0]
        return min(non_zero_values) if non_zero_values else None

    @staticmethod
    def get_chunks_data_by_dict(
        idfy_not_null_dict: dict,
        chunk_size: int,
        is_use_small_chunks: bool = False,
        small_chunk_size: int = 10
    ) -> list[dict[Any, Any]]:
        """Делит СЛОВАРЬ с данными на чанки с данными"""
        if chunk_size <= 0:
            raise ValueError("Размер чанка должен быть положительным")

        if is_use_small_chunks:
            chunk_size = small_chunk_size

        items = list(idfy_not_null_dict.items())
        return [
            dict(items[i:i + chunk_size])
            for i in range(0, len(items), chunk_size)
        ]

    @staticmethod
    def split_into_chunks(big_list, chunk_size=100):
        """
        Разделение большого СПИСКА на меньшие кусочки фиксированного размера.

        :param big_list: Основной список для деления.
        :param chunk_size: Размер каждого фрагмента (по умолчанию 100).
        :return: Список фрагментов.
        """
        return [big_list[i:i + chunk_size] for i in range(0, len(big_list), chunk_size)]

    @staticmethod
    def filter_list(a: list, b: list) -> list:
        """
        Оставляет в списке A только элементы, которых нет в списке B

        :param a: Основной список (будет изменен)
        :param b: Список элементов для исключения
        :return: Отфильтрованный список A
        """
        # Преобразуем в множества для быстрого поиска
        set_b = set(b)
        # Оставляем только элементы не из B
        return [x for x in a if x not in set_b]

    @staticmethod
    def get_field_structure(field: str):
        if AsisTobe.ASIS in field:
            return {
                "name": field.replace(f"_{AsisTobe.ASIS}", ""),
                "mod": AsisTobe.ASIS
            }
        elif AsisTobe.TOBE in field:
            return {
                "name": field.replace(f"_{AsisTobe.TOBE}", ""),
                "mod": AsisTobe.TOBE
            }
        else:
            return {
                "name": field,
                "mod": None
            }

    @staticmethod
    def num_tokens_from_messages(messages, model="gpt-4o-mini"):
        """Return the number of tokens used by a list of messages."""
        try:
            encoding = tiktoken.encoding_for_model(model)
        except KeyError:
            print("Warning: model not found. Using o200k_base encoding.")
            encoding = tiktoken.get_encoding("o200k_base")
        if model in {
            "gpt-3.5-turbo-0125",
            "gpt-4-0314",
            "gpt-4-32k-0314",
            "gpt-4-0613",
            "gpt-4-32k-0613",
            "gpt-4o-mini-2024-07-18",
            "gpt-4o-2024-08-06"
        }:
            tokens_per_message = 3
            tokens_per_name = 1
        elif "gpt-3.5-turbo" in model:
            print(
                "Warning: gpt-3.5-turbo may update over time. Returning num tokens assuming gpt-3.5-turbo-0125."
            )
            return Tools.num_tokens_from_messages(messages, model="gpt-3.5-turbo-0125")
        elif "gpt-4o-mini" in model:
            print(
                "Warning: gpt-4o-mini may update over time. Returning num tokens assuming gpt-4o-mini-2024-07-18."
            )
            return Tools.num_tokens_from_messages(messages, model="gpt-4o-mini-2024-07-18")
        elif "gpt-4o" in model:
            print(
                "Warning: gpt-4o and gpt-4o-mini may update over time. Returning num tokens assuming gpt-4o-2024-08-06."
            )
            return Tools.num_tokens_from_messages(messages, model="gpt-4o-2024-08-06")
        elif "gpt-4" in model:
            print("Warning: gpt-4 may update over time. Returning num tokens assuming gpt-4-0613.")
            return Tools.num_tokens_from_messages(messages, model="gpt-4-0613")
        else:
            raise NotImplementedError(
                f"""num_tokens_from_messages() is not implemented for model {model}."""
            )
        num_tokens = 0
        for message in messages:
            num_tokens += tokens_per_message
            for key, value in message.items():
                num_tokens += len(encoding.encode(value))
                if key == "name":
                    num_tokens += tokens_per_name
        num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens

    @staticmethod
    def remove_duplicate_words(text: str) -> str:
        """
        Удаляет повторяющиеся слова в строке, сохраняя порядок и регистр первого вхождения.

        :param text: Исходная строка
        :return: Строка без повторяющихся слов
        """
        seen_words = set()
        result = []

        for word in text.split():
            # Приводим к нижнему регистру только для проверки дубликатов
            lower_word = word.lower()

            if lower_word not in seen_words:
                seen_words.add(lower_word)
                result.append(word)  # Сохраняем оригинальное слово с регистром

        return ' '.join(result)

    @staticmethod
    def clear_directory_contents(path: Path, with_dir: bool = False) -> None:
        """
        Полностью очищает содержимое указанной папки (включая подпапки и файлы).
        Не удаляет саму папку, только её содержимое.

        Args:
            path: Путь к папке, которую нужно очистить (Path объект)

        Raises:
            ValueError: Если переданный путь не является директорией
            OSError: При ошибках удаления файлов
            :param path:
            :param with_dir: вместе с папкой
        """
        if not path.is_dir():
            raise ValueError(f"Путь {path} не является директорией")

        # Удаляем все содержимое рекурсивно
        for item in path.iterdir():
            try:
                if item.is_file() or item.is_symlink():
                    item.unlink()  # Удаляем файлы и симлинки
                elif item.is_dir():
                    shutil.rmtree(item)  # Рекурсивно удаляем поддиректории
            except Exception as e:
                raise OSError(f"Ошибка при удалении {item}: {e}")

    @staticmethod
    def parse_filename(filename):
        """
        Парсит имя файла формата "2417_Лестницы_и_стремянки_20250529_на_разметку.xlsx"
        и возвращает категорию и название категории.

        Args:
            filename (str): Имя файла для парсинга

        Returns:
            tuple: (категория, название_категории)
        """
        # Удаляем расширение файла
        name_without_ext = filename.split('.')[0]

        # Разбиваем по символу подчеркивания
        parts = name_without_ext.split('_')

        # Категория - первая часть
        # parts[0] = "123dwd"
        category = parts[0]
        if not category.isdigit():
            raise ValueError("Категория должна состоять из цифр")

        category = int(category)

        # Название категории - все части между категорией и датой
        # Ищем индекс первой части, содержащей дату (формат YYYYMMDD)
        date_pattern = re.compile(r'^\d{8}$')
        date_index = None

        for i, part in enumerate(parts):
            if date_pattern.match(part):
                date_index = i
                break

        if date_index is None:
            # Если дата не найдена, берем все части после категории
            name_parts = parts[1:]
        else:
            name_parts = parts[1:date_index]

        # Собираем название категории, заменяя подчеркивания на пробелы
        category_name = ' '.join(name_parts)

        return category, category_name

    @staticmethod
    def idfy_by_field(
        data: dict | list,
        key_field: str,
        save_repeats: bool = True,
        repeat_char: str = "_"
    ):
        """
        Метод проводит словаризацию словаря/списка по данным из поля
        :param repeat_char: символ, который ставится после идентификатора поля, если такой элемент уже есть
        :param save_repeats: сохранять повторы?
        :param data:
        :param key_field: поле, из которого будет браться ключ для словаря
        :return:
        """
        if not isinstance(data, list):
            raise TypeError("Данные должны быть списком")

        result: dict[Any, Any] = {}
        item: dict
        for item in data:
            if not isinstance(item, dict):
                raise TypeError("Элемент должен быть словарём")
            if key_field not in item:
                raise AttributeError(f"Поле «{key_field}» отсутствует в элементе")

            key_value: Optional[Any] = item.get(key_field)
            if not key_value:
                raise ValueError(f"В качестве ключа должен быть хэшируемое значение: {key_value}")

            while True:
                if key_value not in result:
                    result[key_value] = item
                    break
                if save_repeats:
                    key_value = f"{repeat_char}{key_value}"
        return result

    @staticmethod
    def is_valid_json(s):
        """Проверка строки на валидное json-содержимое"""
        try:
            json.loads(s)
            return True
        except (json.JSONDecodeError, TypeError, ValueError):
            return False

    @staticmethod
    def str2json(text: str) -> Optional[list | dict]:
        try:
            json_data = json.loads(text)
        except json.JSONDecodeError:
            return None
        return json_data

    @staticmethod
    def get_clean_words(words: list[str]) -> list[str]:
        """Берет список строк, делает их lower"""
        words_lower = [word.lower() for word in words]
        return [Tools.clean_field(word) for word in words_lower]

    @staticmethod
    def get_fields_pairs(fields: list[str]) -> dict[str, dict[str, str]]:
        if not len(fields):
            raise IndexError("Список полей для парирации не может быть пустым")
        result: dict[str, dict[str, str]] = {}
        field: str
        for field in fields:
            result[field] = {
                AsisTobe.ASIS: f"{field}_{AsisTobe.ASIS}",
                AsisTobe.TOBE: f"{field}_{AsisTobe.TOBE}",
            }
        return result

    @staticmethod
    def get_word_by_index(
        words_list: list[str],
        words_indexes: list[int],
        *,
        ignore_errors: bool = True,
        unique_only: bool = True
    ) -> list[str]:
        """
        Усовершенствованная версия с поддержкой уникальных слов без потери порядка.
        ...
        """
        if not words_list or not words_indexes:
            return []

        result = []
        seen_words = set()

        for i in words_indexes:
            if 0 <= i < len(words_list):
                word = words_list[i]
                if unique_only:
                    if word not in seen_words:
                        result.append(word)
                        seen_words.add(word)
                else:
                    result.append(word)
            elif not ignore_errors:
                raise IndexError(f"Индекс {i} выходит за пределы списка (длина: {len(words_list)})")

        return result



def value_to_bool(value: any) -> bool:
    """
    Преобразует значение в bool по расширенным правилам.

    True: 'true', '1', 1, 'yes', 'on', '+', 'ok', 'да', 'вкл', True
    False: 'false', '0', 0, 'no', 'off', '-', 'нет', 'выкл', False, None, '', ' '

    :param value: Любое значение
    :return: bool
    :raises: ValueError при неоднозначных строках (если включить strict)
    """
    if value is None:
        return False

    if isinstance(value, bool):
        return value

    if isinstance(value, (int, float)):
        return bool(value)

    if not isinstance(value, str):
        # Если не строка, не число и не bool — пытаемся привести к строке
        try:
            value = str(value).strip()
        except Exception:
            return False

    # Приводим к строке и очищаем
    value = value.strip()

    if not value:
        return False

    # Нормализуем к нижнему регистру
    normalized = value.lower()

    # Списки значений
    true_values = {
        'true', '1', 'yes', 'on', 'ok', 'да', 'вкл', 'y', 't', '+', 'enable', 'enabled'
    }
    false_values = {
        'false', '0', 'no', 'off', 'нет', 'выкл', 'n', 'f', '-', 'disable', 'disabled', 'null',
        'none'
    }

    if normalized in true_values:
        return True
    if normalized in false_values:
        return False

    # Если значение не распознано
    # Можно выбросить ошибку или вернуть False (по умолчанию — False)
    return False



"""ЯДРО"""

import warnings
from pathlib import Path
from typing import Optional

import pandas as pd
from deprecated import deprecated
from pandas import DataFrame, read_excel

from ..configs.column_strings import ColumnStrings
from ..configs.system import System
from ..logger.core import log, logerr, loginf, logsuc
from ..pretty_print.core import pretty_print
from ..reader_controller.exceptions import ContentLengthError, WrongSheetListError
from ..reader_controller.types import DFType, IdfyGoods, XLSGood

print = log


class ReaderController:
    """Контроллер для управления XLS(X) файлами"""

    def __init__(
        self,
        file_path: Path,
        file_output: Path,
        is_new: bool = False,
        skip_rows: int = 0,
        debug: bool = False,
    ):
        self._is_new: bool = is_new
        self._file_path: Path = file_path
        self.check_exists()
        self._file_output: Path = file_output
        self._skip_rows: int = skip_rows
        self._dataframe: list[dict] = []
        self._idfy_dataframe: dict = {}
        self._local_idfy_dataframe: dict = {}
        self._debug: bool = debug

        self.check_writable()

    def read_data(self, sheet_name: Optional[str] = None):
        """Чтение файла"""
        self._dataframe = self._sheet_read(sheet_name=sheet_name)

    def filtered_data(self):
        """
        Фильтрованными по 0 значениями - нули не включать
        исходник - 1000
        на обработку ai_asis - 900
        :return:
        """
        return self._dataframe

    @classmethod
    def get_not_nulled_data(cls, data_dict: dict):
        """Получает ненульные строки"""
        not_nulled_dict = {}
        data: dict
        for index, data in data_dict.items():
            # print(f"{index=} ({type(index)})")
            if data.get(ColumnStrings.DATA_ENTITY_TOBE) in System.NULLED:
                continue
            not_nulled_dict[index] = data
        return not_nulled_dict

    @classmethod
    def get_data_for_remark(cls, data_dict: dict):
        """Получаем данные для переразметки"""
        result = {}
        data_item: dict
        for local_id, data_item in data_dict.items():
            remark: Optional[str] = data_item.get(ColumnStrings.RMK)
            if remark is None:
                continue

            if str(remark) == "1":
                result[local_id] = data_item
                continue

        return result

    @classmethod
    def data_with_only_source_name(cls, data_list: DFType):
        """Получить только local_id и source_name"""
        new_list: DFType = []
        item: XLSGood
        for item in data_list:
            new_list.append(
                {
                    ColumnStrings.DATA_LOCAL_ID: item.get(ColumnStrings.DATA_LOCAL_ID, ''),
                    ColumnStrings.DATA_SOURCE_NAME: item.get(ColumnStrings.DATA_SOURCE_NAME, ''),
                }
            )
        return new_list

    @classmethod
    def valid_no_empty_source_name(cls, data_list: DFType):
        """Проверяет сурснеймы на пустоту"""
        result = True
        item: XLSGood
        for item in data_list:
            source_name: Optional[str] = item.get(ColumnStrings.DATA_SOURCE_NAME)
            if not source_name or source_name in System.FULL_SKIP:
                result = False
                break
        return result

    def _remove_first_sheet_and_rewrite(self):
        """
        Удаляет первый лист из Excel-файла и пересохраняет файл без него.
        Использует pandas и openpyxl для записи.
        """
        sheet_names = self._get_excel_sheet_names()

        if len(sheet_names) <= 1:
            raise ValueError("Невозможно удалить единственный лист")

        try:
            # Читаем все листы, кроме первого
            new_sheets = {}
            for name in sheet_names[1:]:
                df = read_excel(self._file_path, sheet_name=name)
                new_sheets[name] = df.fillna('').map(self._convert_to_string, na_action='ignore')

            # Перезаписываем файл, используя существующую логику записи
            # Временно меняем путь, чтобы не сломать update_file
            temp_df = DataFrame()  # заглушка
            temp_df.to_excel(
                self._file_path, index=False, sheet_name=sheet_names[1]
            )  # костыль для очистки

            with pd.ExcelWriter(self._file_path, engine='openpyxl') as writer:
                for sheet_name, data in new_sheets.items():
                    data.to_excel(writer, sheet_name=sheet_name, index=False)

            if self._debug: logsuc(f"Лист '{sheet_names[0]}' удалён. Файл пересохранён.")
        except PermissionError:
            raise PermissionError("Файл открыт в Excel — невозможно пересохранить.")
        except Exception as e:
            logerr(f"Ошибка при удалении листа: {e}")
            raise

    def _get_excel_sheet_names(self) -> list[str]:
        """
        Возвращает список имён листов в Excel-файле.
        Использует pandas.ExcelFile для безопасного чтения метаданных.
        """
        try:
            with pd.ExcelFile(self._file_path) as xls:
                return xls.sheet_names
        except Exception as e:
            logerr(f"Не удалось прочитать структуру Excel-файла: {e}")
            raise ValueError("Ошибка при чтении списка листов") from e

    def check_writable(self, same_file: bool = True):
        """
        Проверить, не открыт ли файл
        :param same_file:
        :return:
        """
        check_writable: bool = self.is_file_writable(same_file)
        if not check_writable:
            raise PermissionError("ФАЙЛ ОТКРЫТ И НЕДОСТУПЕН ДЛЯ ЗАПИСИ")
        if self._debug: logsuc("Файл с данными существует")

    def check_exists(self):
        """
        Проверка файла на существование
        :return:
        """
        if not self._is_new:
            if not self._file_path.exists():
                raise FileNotFoundError("Файл с данными не найден")
            if self._debug: logsuc("Файл с данными доступен для записи")
        else:
            if self._debug: logsuc("Файл является новым и ожидает записи")

    def is_file_writable(self, same_file: bool = True) -> bool:
        """
        INNER-METHOD
        Проверяет, доступен ли файл для записи.
        Возвращает True если файл доступен для записи, False если заблокирован.
        """
        file: Path = self._file_path
        if not same_file:
            file = self._file_output

        if not file.exists():
            return True  # Файл не существует - можно создавать

        try:
            # Пытаемся открыть файл в режиме append для проверки блокировки
            with open(file, 'a'):
                pass
            return True
        except PermissionError:
            return False
        except Exception as e:
            print(f"Неожиданная ошибка при проверке файла: {e}")
            return False

    @property
    def rows_total(self):
        """Вообще все строки"""
        return self.rows_data + 1

    @property
    def rows_data(self):
        """Только строки с данными без заголовков"""
        return len(self._dataframe)

    def get_sheet_names(self):
        """Имена листов"""
        return self._get_excel_sheet_names()

    def _sheet_read(self, sheet_name: Optional[str] = None) -> DFType:
        """Считывает все данные в словарь с преобразованием всех значений в строки"""
        if sheet_name is None:
            dataframe = read_excel(self._file_path)
        else:
            try:
                dataframe = read_excel(self._file_path, sheet_name=sheet_name)
            except Exception as e:
                raise ValueError(e)

        # Заменяем NaN на пустую строку
        dataframe = dataframe.fillna('')

        # Обрабатываем каждую ячейку: приводим к строке, убираем .0 у чисел
        for col in dataframe.columns:
            dataframe[col] = dataframe[col].map(self._convert_to_string)

        return dataframe.to_dict('records')

    @staticmethod
    def _convert_to_string(value) -> str:
        """Преобразует значение в строку, удаляя .0 у float, если это целое число"""
        if isinstance(value, float):
            if value.is_integer():
                return str(int(value))
            return str(value)
        elif isinstance(value, int):
            return str(value)
        elif isinstance(value, str):
            return value.strip()
        return str(value)

    def get_data(self, sheet_name: Optional[str] = None):
        """Получение данных если их нет"""
        if not len(self._dataframe):
            self.read_data(sheet_name=sheet_name)
        return self._dataframe

    def perfect_data(self, data_list: DFType):
        """Только не нульные строки"""
        data = self.local_idfy(data_list)
        return self.get_not_nulled_data(data)

    @deprecated("ДАННЫЙ МЕТОД НЕ ИСПОЛЬЗУЕТСЯ ИЗ-ЗА УСТАРЕВШЕЙ МЕТОДИКИ ПРИСВОЕНИЯ LOCAL_ID")
    def idfy_data(self, data_list: XLSGood):
        """
        Присваивает нумерацию строк в зависимости от ПОРЯДКОВОГО НОМЕРА строки

        :param data_list:
        :return:
        """

        warnings.warn(
            "ReaderController.idfy_data() устарел и будет удалён в будущих версиях. "
            "Используйте local_idfy() или idfy_to_dataframe().",
            DeprecationWarning,
            stacklevel=2
        )
        self._idfy_dataframe = {int(idx) + 1: item for idx, item in enumerate(data_list)}
        return self._idfy_dataframe

    def local_idfy(self, data_list: DFType):
        """Присваивает номера в зависимости от РАНЕЕ ПРИСВОЕННОГО номера - по сути ID"""
        item: dict
        self._local_idfy_dataframe = {
            int(item.get(ColumnStrings.DATA_LOCAL_ID)): item for item in data_list
        }
        return self._local_idfy_dataframe

    def update_file(self, same_file=True):
        """
        Обновляет Excel-файл по self._file_path данными из data_list.
        :param same_file:
        """
        check_writable: bool = self.is_file_writable(same_file=same_file)
        if not check_writable:
            raise PermissionError("ФАЙЛ НЕ ДОСТУПЕН ДЛЯ ЗАПИСИ, ОТКРЫТ!")
        df = DataFrame(self._dataframe)
        if not same_file:
            df.to_excel(self._file_output, index=False)
        else:
            df.to_excel(self._file_path, index=False)
        return True

    @classmethod
    def response_validator_substr(
        cls,
        data_response: list[dict],
        validated_fields: list[str],
        is_ignore: bool = False,
        filled: bool = False
    ):
        """

        :param data_response:
        :param validated_fields:
        :param is_ignore:
        :param filled:
        :return:
        """
        # TODO ПРОВЕРИТЬ СОВПАДЕНИЕ ID ТОГО, ЧТО ОТПРАВИЛ И ТОГО ЧТО ВЕРНУЛА НЕЙРОНКА - raise
        result = []
        for data in data_response:
            if 'ID' not in data:
                logerr("Не найден ID строки")
                pretty_print(data, title=f"data", m2d=False)
                raise

            for field in validated_fields:
                if field not in data:
                    logerr(f"Не найдено поле {field}")
                    pretty_print(data, title=f"data", m2d=False)
                    raise

            if not is_ignore:
                for field in validated_fields:
                    search_value: Optional[str] = data.get(field)
                    if not search_value:
                        # logerr(f"Данные в поле {field} не обнаружены")
                        # pretty_print(data, title=f"data", m2d=False)
                        # raise
                        data[field] = "no_data"
                        search_value = data[field]

                    if search_value not in ["wrong_cat", "no_data"]:
                        source_data: Optional[str] = data.get("source_name")
                        if not source_data:
                            logerr(f"Источник в ответе не обнаружен")
                            pretty_print(data, title=f"data", m2d=False)
                            raise

                        if search_value.lower() not in source_data.lower():
                            logerr(f"Подстрока {field}={search_value} НЕ НАЙДЕНА в source_name")
                            pretty_print(data, title=f"data", m2d=False)
                            raise

            result.append(data)

        return result

    def override_dataframe(self, data_list, is_hard: bool = False):
        """
        Перезапись датафрейма
        :param data_list:
        :param is_hard: ПРИНУДИТЕЛЬНО БЕЗ ПРОВЕРОК
        :return:
        """
        if not len(data_list):
            raise ValueError("СОДЕРЖИМОЕ ПУСТОЕ, НЕ МОГУ ПЕРЕЗАПИСАТЬ ДАТАФРЕЙМ")
        if not is_hard:
            if len(self._dataframe) != len(data_list):
                raise ValueError("СОДЕРЖИМОЕ ПО КОЛИЧЕСТВУ РАЗНОЕ, НЕ МОГУ ПЕРЕЗАПИСАТЬ ДАТАФРЕЙМ")
        self._dataframe = data_list

    def _rewrite_dataframe_by_local(self):
        len_local = len(self._local_idfy_dataframe)
        len_main = len(self._dataframe)
        if len_local != len_main:
            raise ValueError("Датафреймы не равны для обновления")
        result = []
        for _, data_item in self._local_idfy_dataframe.items():
            result.append(data_item)
        return result

    def update_dataframe_from_updated_dataframe(
        self,
        updated_dataframe,
        updated_fields: list,
        field_id: str = "ID",
    ) -> Optional[bool]:
        """Обновляет главный датафрейм, используя данные обновлённого датафрейма"""
        if not len(updated_dataframe):
            logerr("Пустой обновлённый датафрейм")
            return None
        if not len(updated_fields):
            logerr("Не указаны обновляемые поля")
            return None

        # pretty_print(self._local_idfy_dataframe, title=f"self._local_idfy_dataframe", m2d=False)
        for u_id, u_data in updated_dataframe.items():
            ID: int = int(u_id)
            if ID not in self._local_idfy_dataframe:
                print(f"{ID=}")
                print(f"{u_id=}")
                pretty_print(u_data, title=f"u_data", m2d=False)
                # self.show_local_idfy_dataframe()
                raise IndexError("ID НЕ НАЙДЕН В LOCAL_IDFY")

            for field in updated_fields:
                if field not in u_data:
                    print(f"{ID=}")
                    print(f"{field=}")
                    pretty_print(u_data, title=f"u_data", m2d=False)
                    raise IndexError("FIELD НЕ НАЙДЕН В U_DATA")

                local_id_data_item: dict = self._local_idfy_dataframe.get(ID)
                if field not in local_id_data_item:
                    print(f"{ID=}")
                    print(f"{field=}")
                    pretty_print(local_id_data_item, title=f"local_id_data_item", m2d=False)
                    raise IndexError("FIELD НЕ НАЙДЕН В local_id_data_item")

                self._local_idfy_dataframe[ID][field] = u_data[field]
        self._dataframe = self._rewrite_dataframe_by_local()
        return True

    def update_dataframe(
        self,
        data_response: list[dict],
        required_fields: list[str],
        primary_key: Optional[str] = None,
    ):
        """
        Старый метод обновления датафрейма
        Обновляет idfy_dataframe из приходящего list[dict]
        :param data_response:
        :param required_fields:
        :param primary_key:
        :return:
        """
        if primary_key is None:
            primary_key = System.ID

        if not len(data_response):
            loginf("Нет данных для изменений")
            return self._dataframe
        data: dict
        for data in data_response:
            ID: Optional[str] = data.get(primary_key, None)
            if not ID:
                pretty_print(data, title=f"data", m2d=False)
                raise ValueError("ID НЕ УКАЗАН В RESPONSE")
            index: Optional[int] = int(ID)
            if not index:
                logerr("INDEX не определён")
                pretty_print(data, title=f"data", m2d=False)
                raise
            if index not in self._idfy_dataframe:
                logerr("INDEX не найден")
                pretty_print(data, title=f"data", m2d=False)
                raise

            for field in required_fields:
                self._idfy_dataframe[index][field] = data.get(field)

    def modify_dataframe(self):
        """
        Применение изменений из idfy в основной dataframe
        :return:
        """
        self._dataframe = list(self._idfy_dataframe.values())
        return self._dataframe

    @classmethod
    def _check_field_in_data_item(cls, data_item: dict, field: str) -> bool:
        """Проверяет поле field в data_item"""
        return field in data_item

    def _right_sheet(self, data_item: dict) -> bool:
        """Поиск SOURCE_NAME и URL"""
        loginf("Right sheet?..")
        if (
            self._check_field_in_data_item(data_item, ColumnStrings.DATA_SOURCE_NAME)
            and
            self._check_field_in_data_item(data_item, ColumnStrings.DATA_URL)
        ):
            return True
        # raise WrongDocumentErr("ПЕРВЫЙ ЛИСТ ДОКУМЕНТА НЕ ЯВЛЯЕТСЯ ЛИСТОМ С ДАННЫМИ")
        return False

    def cycle_right_sheet(self) -> None:
        """
        Циклически проверяет первый лист Excel-файла.
        Если не содержит нужные поля (source_name, url) — удаляет его и повторяет.
        Останавливается, когда находит подходящий лист или остаётся один лист.
        """
        loginf("Запуск поиска подходящего листа...")

        while True:
            sheet_names = self._get_excel_sheet_names()
            loginf(f"Остались листы: {sheet_names}")

            # Пытаемся прочитать первый лист
            try:
                # Временно читаем только первый лист, как делает _sheet_read
                temp_df = read_excel(self._file_path, sheet_name=sheet_names[0])
                temp_df = temp_df.fillna('')
                for col in temp_df.columns:
                    temp_df[col] = temp_df[col].map(self._convert_to_string)
                records = temp_df.to_dict('records')
            except Exception as e:
                logerr(f"Не удалось прочитать лист '{sheet_names[0]}': {e}")
                if len(sheet_names) == 1:
                    raise WrongSheetListError("Единственный лист повреждён или пуст") from e
                else:
                    # Пробуем следующий
                    self._remove_first_sheet_and_rewrite()
                    continue

            # Проверяем, подходит ли первый элемент
            if records and self._right_sheet(records[0]):
                logsuc(f"Найден подходящий лист: '{sheet_names[0]}'")
                return

            # Лист не подходит
            if len(sheet_names) == 1:
                logerr(f"Последний лист '{sheet_names[0]}' не содержит source_name и url")
                raise WrongSheetListError("Ни один лист не содержит обязательных полей")

            # Удаляем первый лист и повторяем
            self._remove_first_sheet_and_rewrite()

            # Сбрасываем внутреннее состояние, чтобы при следующем read_data() был перечитан файл
            self._dataframe = []

    def check_local_id(self, find_it: bool = True):
        """Проверяет, присутствует ли в файле local_id"""
        if not self._dataframe:
            self.read_data()
        if not self._dataframe:
            raise ContentLengthError("Нет данных")
        data_item: XLSGood = self._dataframe[0]
        if not data_item.keys():
            raise ContentLengthError("Нет данных в позиции")

        self.cycle_right_sheet()
        if find_it:
            return self._check_field_in_data_item(data_item, ColumnStrings.DATA_LOCAL_ID)

        return True

    def process_local_idfying(self, field: str = ColumnStrings.DATA_LOCAL_ID):
        """Проставляем {field} внутрь файла"""
        log("Добавляю LOCAL_ID...")
        if not self._dataframe:
            self.read_data()

        data: DFType = self.get_data()
        data_item: XLSGood
        for index, data_item in enumerate(data):
            data_idfy: dict = {
                "local_id": index + 1,
                **data_item
            }
            data[index] = data_idfy

        self.override_dataframe(data)
        self.update_file()
        logsuc("В файл добавлен LOCAL_ID")

    def get_asis_fields(self):
        """Получает все asis-столбцы"""
        if not self._dataframe:
            self.read_data()
        if not self._dataframe:
            raise ContentLengthError("Нет данных")
        data_item: XLSGood = self._dataframe[0]
        if not data_item.keys():
            raise AttributeError("Нет данных в итеме")
        result: list[str] = []
        for field, _ in data_item.items():
            if "_asis" in field:
                result.append(field)
        return sorted(result)

    def get_tobe_fields(self):
        """Получает все tobe-столбцы"""
        if not self._dataframe:
            self.read_data()
        if not self._dataframe:
            raise ContentLengthError("Нет данных")
        data_item: XLSGood = self._dataframe[0]
        if not data_item.keys():
            raise AttributeError("Нет данных в итеме")
        result: list[str] = []
        for field, _ in data_item.items():
            if "_tobe" in field:
                result.append(field)
        return sorted(result)

    def get_all_fields(self):
        """Получает все столбцы"""
        return sorted(
            [
                *self.get_asis_fields(),
                *self.get_tobe_fields()
            ]
        )

    def idfy_to_dataframe(self, idfy_data: IdfyGoods):
        """Преобразовать idfy-данные в DFType (датафрейм)"""
        self._dataframe = [data_item for _, data_item in idfy_data.items()]
        return self._dataframe

    def save_to_csv(self, same_file=True):
        """
        Сохранить в формате CSV
        Необходимо в _file_path|_file_output прописывать .csv
        """
        df = DataFrame(self._dataframe)
        if not same_file:
            df.to_csv(self._file_output, index=False)
        else:
            df.to_csv(self._file_path, index=False)
        return True

    def rename(self, new_name: str, same_file: bool = True) -> bool:
        """Переименовывает файл."""
        # Определяем путь и директорию назначения
        file_path = self._file_path if same_file else self._file_output
        target_dir = file_path.parent

        # Формируем новое имя с сохранением расширения
        new_path = target_dir / f"{new_name}{file_path.suffix}"

        # Переименовываем
        file_path.rename(new_path)

        # Обновляем путь в объекте
        if same_file:
            self._file_path = new_path
        else:
            self._file_output = new_path

        return True

    def show_local_idfy_dataframe(self):
        """Показать local_idfy"""
        pretty_print(self._local_idfy_dataframe, title=f"self._local_idfy_dataframe", m2d=False)

    def show_idfy_dataframe(self):
        """Показать idfy_dataframe"""
        pretty_print(self._idfy_dataframe, title=f"self._idfy_dataframe", m2d=False)



import hashlib
import json
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Generator, List, Optional, Union

try:
    import ijson

    IJSON_AVAILABLE = True
except ImportError:
    IJSON_AVAILABLE = False


class JsonReader:
    """
    Умный читатель и валидатор JSON-файлов.

    Предоставляет безопасный, быстрый и удобный интерфейс для:
     - Валидации синтаксиса JSON
     - Проверки целостности файла
     - Получения метаданных без полной загрузки
     - Ленивой загрузки содержимого
     - Хеширования и контроля изменений

    Поддерживает два режима:
     1. Обычный (`use_stream=False`) — подходит для файлов до ~50 МБ.
     2. Потоковый (`use_stream=True`) — для больших файлов (>50 МБ), требует `ijson`.

    Пример использования:
        reader = JsonReader("data.json")
        if reader.is_valid:
            print(reader.data)  # Данные загружаются только при первом доступе
    """

    def __init__(self, file_path: Union[str, Path], use_stream: bool = False):
        """
        Инициализация читателя JSON.

        Args:
            file_path: Путь к JSON-файлу.
            use_stream: Использовать потоковую валидацию (для больших файлов).
                        Требует установленного пакета `ijson`.

        Raises:
            ValueError: Если use_stream=True, но ijson не установлен.
            FileNotFoundError: Если файл не существует.
        """
        self._file_path: Path = Path(file_path).resolve()
        self._use_stream: bool = use_stream
        self._data: Optional[Any] = None
        self._validation_error: Optional[Exception] = None

        if use_stream and not IJSON_AVAILABLE:
            raise ValueError(
                "Потоковая валидация требует установки 'ijson'. "
                "Выполните: pip install ijson"
            )

        if not self._file_path.exists():
            raise FileNotFoundError(f"Файл не найден: {self._file_path}")

    @property
    def path(self) -> Path:
        """Полный путь к файлу."""
        return self._file_path

    @property
    def name(self) -> str:
        """Имя файла (без пути)."""
        return self._file_path.name

    @property
    def stem(self) -> str:
        """Имя файла без расширения."""
        return self._file_path.stem

    @property
    def suffix(self) -> str:
        """Расширение файла (включая точку)."""
        return self._file_path.suffix

    @property
    def exists(self) -> bool:
        """True, если файл существует."""
        return self._file_path.exists()

    @property
    def is_file(self) -> bool:
        """True, если путь указывает на файл (не директорию)."""
        return self._file_path.is_file()

    @property
    def size_bytes(self) -> int:
        """Размер файла в байтах."""
        return self._file_path.stat().st_size if self.exists else 0

    @property
    def size_mb(self) -> float:
        """Размер файла в мегабайтах (округлён до 2 знаков)."""
        return round(self.size_bytes / (1024 * 1024), 2)

    @property
    def created_at(self) -> Optional[datetime]:
        """Дата создания файла (точная семантика зависит от ОС)."""
        if not self.exists:
            return None
        timestamp = self._file_path.stat().st_ctime
        return datetime.fromtimestamp(timestamp)

    @property
    def modified_at(self) -> Optional[datetime]:
        """Дата последнего изменения файла."""
        if not self.exists:
            return None
        timestamp = self._file_path.stat().st_mtime
        return datetime.fromtimestamp(timestamp)

    @property
    def lines(self) -> int:
        """Примерное количество строк в файле (оценка)."""
        if not self.exists or not self.is_file:
            return 0
        try:
            with self._file_path.open('rb') as f:
                return sum(1 for _ in f)
        except Exception:
            return 0

    @property
    def sha256(self) -> str:
        """SHA-256 хеш содержимого файла."""
        if not self.exists:
            return ""
        hash_sha256 = hashlib.sha256()
        with self._file_path.open('rb') as f:
            for chunk in iter(lambda: f.read(8192), b""):
                hash_sha256.update(chunk)
        return hash_sha256.hexdigest()

    @property
    def data(self) -> Any:
        """
        Загружает и возвращает содержимое JSON-файла.

        Ленивая загрузка: данные читаются только при первом вызове.
        При повторных вызовах используется кеш.

        Returns:
            Словарь, список или примитив (в зависимости от JSON).

        Raises:
            ValueError: Если JSON некорректен или файл недоступен.
            OSError: Ошибка чтения файла.
        """
        if self._data is not None:
            return self._data

        if not self.is_valid:
            raise ValueError(f"Невозможно прочитать файл: {self._validation_error}")

        try:
            text = self._file_path.read_text(encoding='utf-8')
            self._data = json.loads(text)
            return self._data
        except json.JSONDecodeError as e:
            raise ValueError(f"Ошибка парсинга JSON: {e}")
        except UnicodeDecodeError as e:
            raise ValueError(f"Файл не в UTF-8: {e}")
        except Exception as e:
            raise OSError(f"Ошибка чтения файла: {e}")

    @property
    def data_type(self) -> Optional[str]:
        """Тип корневого объекта: 'dict', 'list', 'str' и т.д. None — если ошибка."""
        try:
            data = self.data
            return type(data).__name__
        except Exception:
            return None

    @property
    def length(self) -> Optional[int]:
        """Длина корневого объекта (len), если это list/dict/str. Иначе None."""
        try:
            data = self.data
            if isinstance(data, (list, dict, str)):
                return len(data)
            return None
        except Exception:
            return None

    @property
    def is_valid(self) -> bool:
        """
        Проверяет, является ли файл валидным JSON.

        При первом вызове выполняется полная валидация.
        Результат кешируется.

        Returns:
            True, если файл существует, читаем и содержит валидный JSON.
        """
        if self._validation_error is not None:
            return False
        if not self.exists:
            self._validation_error = FileNotFoundError(f"Файл не найден: {self._file_path}")
            return False
        if not self.is_file:
            self._validation_error = IsADirectoryError(f"Не файл: {self._file_path}")
            return False
        if self.suffix.lower() != '.json':
            self._validation_error = ValueError(f"Ожидался .json: {self._file_path}")
            return False

        try:
            if self._use_stream:
                if not IJSON_AVAILABLE:
                    raise ImportError("ijson не установлен")
                with self._file_path.open('rb') as f:
                    parser = ijson.parse(f)
                    for _ in parser:
                        pass
            else:
                text = self._file_path.read_text(encoding='utf-8')
                data = json.loads(text)
                if not isinstance(data, (dict, list)):
                    raise ValueError(
                        f"Корень должен быть dict/list, получен: {type(data).__name__}"
                    )
            return True
        except (json.JSONDecodeError, ValueError, ImportError, OSError) as e:
            self._validation_error = e
            return False
        except Exception as e:
            self._validation_error = OSError(f"Неожиданная ошибка: {e}")
            return False

    @property
    def validation_error(self) -> Optional[str]:
        """Сообщение об ошибке валидации или None, если всё в порядке."""
        return str(self._validation_error) if self._validation_error else None

    def validate_structure(self, required_keys: Optional[List[str]] = None) -> bool:
        """
        Проверяет структуру JSON: наличие обязательных ключей (если это dict).

        Args:
            required_keys: Список ключей, которые должны присутствовать в корне.

        Returns:
            True, если структура соответствует.

        Raises:
            ValueError: Если файл не валиден или не dict.
        """
        if not self.is_valid:
            raise ValueError(f"Файл не валиден: {self.validation_error}")

        data = self.data
        if not isinstance(data, dict):
            raise ValueError("Требуется JSON-объект (dict)")

        if required_keys:
            return all(key in data for key in required_keys)

        return True

    def iter_items(self, prefix: str = '') -> Generator[Any, None, None]:
        """
        Потоковое чтение элементов (для списков или объектов большого размера).
        Работает только в режиме `use_stream=True`.

        Yields:
            Элементы из массива или пары ключ-значение из объекта.

        Example:
            for item in reader.iter_items():
                print(item)
        """
        if not self._use_stream:
            raise RuntimeError("iter_items() требует use_stream=True")

        if not IJSON_AVAILABLE:
            raise RuntimeError("Требуется ijson для потокового чтения")

        with self._file_path.open('rb') as f:
            if isinstance(self.data, list):
                for item in ijson.items(f, 'item'):
                    yield item
            elif isinstance(self.data, dict):
                for key, value in ijson.kvitems(f, ''):
                    yield key, value
            else:
                yield self.data

    def __repr__(self) -> str:
        return f"<JsonReader valid={self.is_valid} path='{self.name}'>"

    def __bool__(self) -> bool:
        return self.is_valid

    def info(self) -> Dict[str, Any]:
        """
        Возвращает словарь с основными метаданными.

        Returns:
            {
                "path": str,
                "valid": bool,
                "error": str | None,
                "type": str | None,
                "length": int | None,
                "size_mb": float,
                "lines": int,
                "sha256": str,
                "created_at": str | None,
                "modified_at": str | None
            }
        """
        return {
            "path": str(self.path),
            "valid": self.is_valid,
            "error": self.validation_error,
            "type": self.data_type,
            "length": self.length,
            "size_mb": self.size_mb,
            "lines": self.lines,
            "sha256": self.sha256,
            "created_at": self.created_at.isoformat() if self.created_at else None,
            "modified_at": self.modified_at.isoformat() if self.modified_at else None,
        }

    @staticmethod
    def write(
        data: Any,
        file_path: str | Path,
        indent: Optional[int] = 2,
        ensure_ascii: bool = False,
        create_parents: bool = True,
        atomic: bool = True
    ) -> None:
        """
        Записывает данные в JSON-файл. Поддерживает Pydantic, Path, int-keys и др.
        """
        path = Path(file_path).resolve()

        if create_parents:
            path.parent.mkdir(parents=True, exist_ok=True)

        if atomic:
            temp_path = path.with_suffix(f"{path.suffix}.tmp")
            target_path = temp_path
        else:
            target_path = path

        try:
            # Преобразуем данные с помощью custom serializer
            serialized = JsonReader._serialize(data)

            with open(target_path, 'w', encoding='utf-8') as f:
                json.dump(
                    serialized,
                    f,
                    indent=indent,
                    ensure_ascii=ensure_ascii,
                    default=JsonReader._fallback_serializer
                )
                f.write('\n')

            if atomic and temp_path.exists():
                temp_path.replace(path)

        except Exception as e:
            raise OSError(f"Ошибка записи JSON-файла {path}: {e}")

    @staticmethod
    def _fallback_serializer(obj: Any) -> Any:
        """
        Резервный сериализатор для json.dump().
        Должен возвращать сериализуемый тип.
        """
        if isinstance(obj, Path):
            return str(obj.as_posix())
        elif hasattr(obj, 'model_dump'):
            return JsonReader._serialize(obj.model_dump())
        else:
            return repr(obj)

    @staticmethod
    def _serialize(data: Any) -> Any:
        """Рекурсивно сериализует структуру с поддержкой моделей, путей и int-ключей."""
        if hasattr(data, 'model_dump'):  # Pydantic v2
            return JsonReader._serialize(data.model_dump())
        elif isinstance(data, dict):
            return {
                str(k): JsonReader._serialize(v)  # Принудительно str(k)
                for k, v in data.items()
            }
        elif isinstance(data, (list, tuple)):
            return [JsonReader._serialize(item) for item in data]
        elif isinstance(data, Path):
            return str(data.as_posix())  # или просто str(data) — зависит от ОС
        elif isinstance(data, (int, float, str, bool)) or data is None:
            return data
        elif isinstance(data, set):
            return JsonReader._serialize(list(data))
        else:
            return repr(data)  # fallback

    @staticmethod
    def write_compact(data: Any, file_path: str | Path) -> None:
        """Быстрая компактная запись без отступов."""
        JsonReader.write(data, file_path, indent=None, ensure_ascii=False)

    @staticmethod
    def write_pretty(data: Any, file_path: str | Path) -> None:
        """Красивая форматированная запись."""
        JsonReader.write(data, file_path, indent=2, ensure_ascii=False)



"""ЯДРО"""

import sys
from typing import Any, Callable

from loguru import logger

logger.remove()

# Добавляем новый обработчик с цветами как в стандартном debug
logger.add(
    sys.stderr,
    format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> - <level>{message}</level>",
    level="DEBUG",
    filter=lambda record: record["level"].name == "DEBUG",
    colorize=True  # Включаем цветной вывод
)
logger.add(
    sys.stderr,
    format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> - <level>{message}</level>",
    level="ERROR",
    filter=lambda record: record["level"].name == "ERROR",
    colorize=True  # Включаем цветной вывод
)
logger.add(
    sys.stderr,
    format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> - <level>{message}</level>",
    level="INFO",
    filter=lambda record: record["level"].name == "INFO",
    colorize=True  # Включаем цветной вывод
)
logger.add(
    sys.stderr,
    format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> - <level>{message}</level>",
    level="SUCCESS",
    filter=lambda record: record["level"].name == "SUCCESS",
    colorize=True  # Включаем цветной вывод
)
log = logger.debug
logerr = logger.error
loginf: Callable[[str, tuple[Any, ...], dict[str, Any]], None] | Callable[[Any], None] = logger.info
logsuc = logger.success
